package infopipes.pipes;

import infopipes.utils.Channel;
import infopipes.utils.PipePort;
import infopipes.utils.PipePortSet;
import infopipes.utils.Polarity;
import infopipes.utils.PolarityChoice;
import infopipes.utils.PolarityChoiceMap;
import infopipes.utils.PolarityConfig;
import infopipes.utils.PolarityMap;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class CompositeFactory {
	/**
	 * Mapping from an internal pipe name to a logical pipe instance.
	 */
	Map<String, LogicalPipe> pipes;
	
	/**
	 * Mapping from an internal pipe name to the valid polarity
	 * configurations of the pipe.  The validity is determined by the
	 * natural configurations of the logical pipe and by the channel
	 * connections between internal pipes.
	 */
	PolarityChoiceMap choices;
	
	/**
	 * The internal pipe inports and the composite interface outports
	 * that had not been connected to a channel.
	 */
	PipePortSet danglingInports;
	
	/**
	 * The internal pipe outports and the composite interface inports
	 * that had not been connected to a channel.
	 */
	PipePortSet danglingOutports;
	
	/**
	 * The channels inside the composite pipe.  These channels can
	 * connect internal pipe ports to each other or to composite
	 * interface ports.
	 */
	Set<Channel> channels;
	
	/**
	 * The internal inteface of the composite pipe.  I represent the
	 * other ends of the external interface ports as ports of the iface
	 * logical pipe (e.g. an external outport is a inport of the iface
	 * pipe) to simplify the treatment of channels.
	 */
	DummyLogicalPipe iface;
	
	/**
	 * The reserved name of the internal interface pipe.
	 */
	public static final String ifaceName = "interface";

	public CompositeFactory() {
		pipes = new HashMap<String, LogicalPipe>();
        choices = new PolarityChoiceMap();
        danglingInports = new PipePortSet();
        danglingOutports = new PipePortSet();
        channels = new HashSet<Channel>();
        iface = new DummyLogicalPipe();
        pipes.put(ifaceName, iface);
        choices.put(ifaceName, iface.validConfigurations());
	}

	/**
	 * Declare an internal pipe in the composite pipe. 
	 * @param name defines what the new pipe should be called
	 * @param op specifies the "type" of the new internal pipe
	 */
	public void addPipe(String name, LogicalPipe op) {
		if (pipes.containsKey(name))
			throw new RuntimeException();
		danglingInports.addPortSpecs(name, op.getInports());
		danglingOutports.addPortSpecs(name, op.getOutports());
		choices.put(name, op.validConfigurations());
		pipes.put(name, op);
	}

	/**
	 * Declare an interface inport for the composite pipe.
	 * @param ident is the name for the inport
	 */
	public void newInport(String ident) {
		iface.newOutport(ident);
		danglingOutports.add(new PipePort(ifaceName, ident));
		choices.put(ifaceName, iface.validConfigurations());
	}
	
	/**
	 * Declare an interface outport for the composite pipe.
	 * @param ident is the name for the inport
	 */
	public void newOutport(String ident) {
		iface.newInport(ident);
		danglingInports.add(new PipePort(ifaceName, ident));
		choices.put(ifaceName, iface.validConfigurations());
	}
	
	/**
	 * Connect internal ports in the composite pipe.  You can connect
	 * an internal pipe outport to an internal pipe inport, an
	 * interface inport to an internal pipe inport, or an internal
	 * pipe outport to an interface outport.  Connecting an interface
	 * inport to an interface outport is disallowed to avoid empty
	 * no-pipe channel loopss.
	 * @param channel is the specification of the channel
	 * @throws a RuntimeException if the channel specification is
	 * invalid (i.e., referring to nonexistent pipes, ports, or that
	 * the channel causes internal polarity conflicts)
	 */
	public void connect(Channel channel) {
		// Sanity checking 1: the channel connects valid ports
		if (channel.src.pipe.equals(ifaceName) &&
			channel.dest.pipe.equals(ifaceName))
			throw new RuntimeException();
		danglingOutports.assertHave(channel.src);
		danglingInports.assertHave(channel.dest);
		channels.add(channel);

		// Sanity checking 2: the channel does not cause a conflict
		PolarityChoiceMap copy;
		copy = new PolarityChoiceMap(choices);
		if (solvePolarity(copy).isEmpty()) {
			channels.remove(channel);
			throw new RuntimeException();
		}
		
		// Commit: store the trimmed pipe polarity choices and remove
		// the connected ports from dangling sets 
		choices = copy;
		danglingOutports.remove(channel.src);
		danglingInports.remove(channel.dest);
	}

	/**
	 * Create a logical pipe according to the declaration of interface
	 * ports, internal pipes, and channels.
	 * @return the instantiated logical pipe
	 */
	public CompositeLogicalPipe instantiate() {
		Map<String, Integer> inports, outports;
		Set<PolarityMap> solution;
		Map<PolarityConfig, PolarityMap> configs;
		
		// Sanity check: all ports have been connected
		danglingInports.assertEmpty();
		danglingOutports.assertEmpty();
		
		// Composite pipe ports are the opposite of the internal
		// interface ones in both data direction and polarity; here we
		// reverse the data direction.
		inports = iface.outports;
		outports = iface.inports;
		
		configs = new HashMap<PolarityConfig, PolarityMap>();
		solution = solvePolarity(choices);
		for (PolarityMap config : solution) {
			
			// Compute the polarity of composite pipe by inverting that
			// of the internal interface pipe, and use the composite
			// pipe polarity to index the internal pipe polarity maps.
			Polarity[] external = config.remove(ifaceName).content;
			for (int index=0; index<external.length; index++)
				external[index] = external[index].invert();
			configs.put(new PolarityConfig(external), config);
		}
		return new CompositeLogicalPipe(inports, outports, configs, pipes, channels);
	}

	/**
	 * This method tries to solve the polarity configurations for
	 * internal pipes and returns a set of results.  It eliminates the
	 * pipe polarity configurations that are known to be invalid based
	 * on the current (possibly partial) channel declarations, and it
	 * makes guesses with backtracking if there is insufficient
	 * constraint to produce a unique solution.  
	 * @param input is the polarity choices of internal pipes
	 * @return the set of internal pipe polarity combinations
	 */
	Set<PolarityMap> solvePolarity(PolarityChoiceMap input) {
		boolean changed;
		String unsolvedPipe;
		int unsolvedPort = 0;
		
		// Eliminate invalid pipe polarity configurations by constraint
		// propagation until we reach a fixpoint (i.e. changed == false)
		do {
			changed = false;
			unsolvedPipe = null;
			
			for (Channel c : channels) {
				PolarityChoice srcConfigs = input.get(c.src.pipe);
				PolarityChoice destConfigs = input.get(c.dest.pipe);
				int srcIndex = pipes.get(c.src.pipe).outportIndex(c.src.ident);
				int destIndex = pipes.get(c.dest.pipe).inportIndex(c.dest.ident);
				Polarity srcPolarity = srcConfigs.testPolarity(srcIndex);
				Polarity destPolarity = destConfigs.testPolarity(destIndex);

				// Case 1: both ports of a channel are unconstrained
				// Action: record the source and pipe for later use
				if (srcPolarity.isUnconstrained() &&
					destPolarity.isUnconstrained()) {
					unsolvedPipe = c.src.pipe;
					unsolvedPort = srcIndex;
				}
				
				// Case 2: only the destination port is constrained
				// Action: propagate polarity to source port
				else if (srcPolarity.isUnconstrained()) {
					srcConfigs.constrainPolarity(srcIndex, destPolarity.invert());
					changed = true;
				}
				
				// Case 3: only the source port is constrained
				// Action: propagate polarity to destination port
				else if (destPolarity.isUnconstrained()) {
					destConfigs.constrainPolarity(destIndex, srcPolarity.invert());
					changed = true;
				}
				
				// Case 4: both ports have the same fixed polarity
				// Action: return an empty set of solutions
				else if (srcPolarity.equals(destPolarity))
					return new HashSet<PolarityMap>();
			}
		} while (changed);
		
		// We have reached a fixpoint... do we have a unique solution?
		// Case 1: no, there is an unconstrained connected port
		// Action: make two guesses and continue propagation
		if (unsolvedPipe != null) {
			Set<PolarityMap> solution;
			PolarityChoiceMap copy;
			
			// Suppose that the unconstrained port is positive...
			copy = new PolarityChoiceMap(input);
			copy.get(unsolvedPipe).constrainPolarity(unsolvedPort, Polarity.P);
			solution = solvePolarity(copy);
			
			// Suppose that the unconstrained port is negative...
			copy = new PolarityChoiceMap(input);
			copy.get(unsolvedPipe).constrainPolarity(unsolvedPort, Polarity.N);
			solution.addAll(solvePolarity(copy));
			return solution;
		}

		// Case 2: yes, there are no unconstrained connected ports
		// Action: compile the result from the unique polarity maps
		else {
			Set<PolarityMap> solution;
			PolarityMap config;
			
			solution = new HashSet<PolarityMap>();
			config = new PolarityMap();
			for (String op : input.keySet())
				config.put(op, new PolarityConfig(input.get(op).one()));
			solution.add(config);
			return solution;
		}
	}
}
