package infopipes.pipes;

import infopipes.utils.PolarityChoice;
import infopipes.utils.PolarityConfig;

import java.util.Set;

/**
 * This interface lists the operations that provides information on
 * a logical pipe and allows programmers to create the corresponding
 * physical pipes.  A logical pipe is an abstract representation of a
 * pipe and does not support data processing functionality.
 * @author Chuan-kai Lin
 */
public interface LogicalPipe {
	/**
	 * This method returns the names of input ports of the pipe. 
	 * @return the set of inport names of the pipe
	 */
	Set<String> getInports();
	
	/**
	 * This method returns the names of output ports of the pipe.
	 * @return the set of outport names of the pipe
	 */
	Set<String> getOutports();
	
	/**
	 * This method returns the integer index of an inport for use
	 * with the push method in a physical pipe. 
	 * @param name is the name of the inport
	 * @return the integer index of the port
	 */
	int inportIndex(String ident);
	
	/**
	 * This method returns the integer index of an outport for use
	 * with the pull method in a physical pipe. 
	 * @param name is the name of the outport
	 * @return the integer index of the port
	 */
	int outportIndex(String ident);
	
	/**
	 * This method returns the port polarity configurations that
	 * a corresponding physical pipe may have.  The programmer should
	 * choose one of these configurations as the argument to the
	 * instantiate method.
	 * @return the polarity configurations a physical pipe may have 
	 */
	PolarityChoice validConfigurations();
	
	/**
	 * This method instantiates a physical pipe for this logical pipe
	 * using the specified polarity configuration.
	 * @param config is the polarity configuration for the physical pipe 
	 * @return the instantiated physical pipe
	 */
	PhysicalPipe instantiate(PolarityConfig config);
}
