package infopipes.pipes;

import infopipes.utils.Port;

/**
 * This interface lists the configuration and data packet communication
 * operations that a physical pipe needs to support.  A physical pipe
 * is a concrete instance of a logical pipe that supports a push-pull
 * data exchange interface under a specific polarity configuration.
 * @author Chuan-kai Lin
 */
public interface PhysicalPipe {
	/**
	 * This method allows the programmer to push a data packet through
	 * a negative inport of the physical pipe.
	 * @param index is the index of the negative inport
	 * @param packet is the data packet to be pushed
	 */
	void push(int index, Object packet);
	
	/**
	 * This method allows the programmer to pull a data packet out
	 * through a negative outport of the physical pipe.
	 * @param index is the index of the negative outport
	 * @return the output data packet
	 */
	Object pull(int index);
	
	/**
	 * This method returns the "push target" for a negative inport.
	 * Using getPush() eliminates the runtime indirection overhead
	 * introduced by composite pipe encapsulation.
	 * @param index is the index of the negative inport
	 * @return the reference to the real push target
	 */
	Port getPush(int index);
	
	/**
	 * This method returns the "pull source" for a negative outport.
	 * Using getPull() eliminates the runtime indirection overhead
	 * introduced by composite pipe encapsulation.
	 * @param index is the index of the negative outport
	 * @return the reference to the real pull source
	 */
	Port getPull(int index);
	
	/**
	 * This method connects the "push source" of the named positive
	 * outport to the given push target.  It links the target to the
	 * real push source so that there would be no runtime overhead due
	 * to composite pipe encapsulation. 
	 * @param index is the index of the positive outport
	 * @param dest is the reference to the push target
	 */
	void linkPush(int index, Port dest);
	
	/**
	 * This method connects the "pull source" of the named positive
	 * inport to the given pull target.  It links the target to the
	 * real pull source so that there would be no runtime overhead due
	 * to composite pipe encapsulation.
	 * @param index is the index of the positive inport
	 * @param src is the reference to the pull target
	 */
	void linkPull(int index, Port src);
	
	/**
	 * Invoking this method indicates that the pipe is now fully
	 * connected to its upstream and downstream pipes.  At this point
	 * the physical pipe can perform validation and/or caching PortGoal
	 * objects in instance variables.
	 */
	void connected();
}
